from os.path import abspath, expanduser
from pathlib import Path
import abc

from landmarkerio import ALL_COLLECTION_ID, dirs_in_dir, FileExt


def load_collection(path):
    with open(str(path), 'rb') as f:
        collection = [l.strip() for l in f.readlines()]
    return [l for l in collection if len(l) > 0]


class CollectionAdapter(object):

    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def collection_ids(self):
        pass

    @abc.abstractmethod
    def collection(self, collection_id):
        pass

    def __str__(self):
        a = ('Serving {} collection(s):'.format(len(self.collection_ids())))
        b = '\n'.join(' - {} ({} assets)'.format(c, len(self.collection(c)))
                      for c in self.collection_ids())
        return a + '\n' + b


class FileCollectionAdapter(CollectionAdapter):

    def __init__(self, collection_dir):
        self.collection_dir = Path(abspath(expanduser(collection_dir)))
        print ('collections: {}'.format(self.collection_dir))
        collection_paths = self.collection_dir.glob('*' + FileExt.collection)
        self._collection = {c.stem: load_collection(c)
                            for c in collection_paths}

    def collection_ids(self):
        return self._collection.keys()

    def collection(self, collection_id):
        return self._collection[collection_id]


class AllCacheCollectionAdapter(CollectionAdapter):

    def __init__(self, cache_dir):
        cache_dir = Path(abspath(expanduser(cache_dir)))
        self._collection = [p.name for p in dirs_in_dir(cache_dir)]
        self._collection_ids = [ALL_COLLECTION_ID]

    def collection_ids(self):
        return self._collection_ids

    def collection(self, collection_id):
        if collection_id == ALL_COLLECTION_ID:
            return self._collection
        else:
            raise ValueError("Only valid collection_id "
                             "is '{}'".format(ALL_COLLECTION_ID))



from enum import Enum


class FileExt(Enum):
    lm = '.ljson'
    template = '.txt'
    collection = '.txt'


class CacheFile(Enum):
    texture = 'texture.jpg'
    image = 'image.json'
    thumbnail = 'thumbnail.jpg'
    mesh_tmp = 'mesh.raw.tmp'
    mesh = 'mesh.raw.gz'


class Server(Enum):
    origin = 'http://www.landmarker.io'
    endpoint = '/api/v1/'


class Endpoints(Enum):
    mode = 'mode'
    images = 'images'
    collections = 'collections'
    landmarks = 'landmarks'
    meshes = 'meshes'
    templates = 'templates'
    points = 'points'
    trilist = 'trilist'
    tcoords = 'tcoords'
    normals = 'normals'
    textures = 'textures'
    thumbnail = 'thumbnails'


class Mimetype(Enum):
    json = 'application/json'
    jpeg = 'image/jpeg'
    binary = 'application/octet-stream'


LM_DIRNAME = 'lmiolandmarks'
TEMPLATE_DINAME = '.lmiotemplates'

ALL_COLLECTION_ID = 'all'

dirs_in_dir = lambda (path): sorted([p for p in path.iterdir() if p.is_dir()])

from ._version import get_versions
__version__ = get_versions()['version']
del get_versions



import itertools
from collections import namedtuple
import os.path as p
from pathlib import Path
from flask import safe_join
import abc

from landmarkerio import TEMPLATE_DINAME, FileExt


Group = namedtuple('Group', ['label', 'n', 'index'])


def parse_group(group):
    # split on \n and strip left and right whitespace.
    x = [l.strip() for l in group.split('\n')]
    label, n_str = x[0].split(' ')
    n = int(n_str)
    index_str = x[1:]
    if len(index_str) == 0:
        return Group(label, n, [])
    index = []
    for i in index_str:
        if ':' in i:
            # User is providing a slice
            start, end = (int(x) for x in i.split(':'))
            index.extend([x, x+1] for x in xrange(start, end))
        else:
            # Just a standard pair of numbers
            index.append([int(j) for j in i.split(' ')])
    indexes = set(itertools.chain.from_iterable(index))
    if min(indexes) < 0 or max(indexes) > n:
        raise ValueError("invalid connectivity")
    return Group(label, n, index)


def group_to_json(group, n_dims):
    group_json = {}
    lms = [{'point': [None] * n_dims}] * group.n
    group_json['landmarks'] = lms
    group_json['connectivity'] = group.index
    group_json['label'] = group.label
    return group_json


def groups_to_json(groups, n_dims):
    lm_json = {
        'version': 1,
        'groups': []
    }
    for g in groups:
        lm_json['groups'].append(group_to_json(g, n_dims))
    return lm_json


def load_template(path, n_dims):
    with open(path, 'rb') as f:
        ta = f.read().strip().split('\n\n')
    groups = [parse_group(g) for g in ta]
    return groups_to_json(groups, n_dims)


class TemplateAdapter(object):
    r"""
    Abstract definition of an adapter that can be passed to app_for_adapter in
    order to generate a legal Flask implementation of landmarker.io's REST API
    for Template retrieval.
    """
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def template_ids(self):
        pass

    @abc.abstractmethod
    def load_template(self, lm_id):
        pass


class FileTemplateAdapter(TemplateAdapter):

    def __init__(self, n_dims, template_dir=None):
        self.n_dims = n_dims
        if template_dir is None:
            # try the user folder
            user_templates = p.expanduser(p.join('~', TEMPLATE_DINAME))
            if p.isdir(user_templates):
                template_dir = user_templates
            else:
                raise ValueError("No template dir provided and "
                                 "{} doesn't exist".format(user_templates))
        self.template_dir = Path(p.abspath(p.expanduser(template_dir)))
        print ('templates: {}'.format(self.template_dir))

    def template_ids(self):
        return [t.stem for t in self.template_paths()]

    def template_paths(self):
        return self.template_dir.glob('*' + FileExt.template)

    def load_template(self, lm_id):
        fp = safe_join(str(self.template_dir), lm_id + FileExt.template)
        return load_template(fp, self.n_dims)


class CachedFileTemplateAdapter(FileTemplateAdapter):

    def __init__(self, n_dims, template_dir=None):
        super(CachedFileTemplateAdapter,
              self).__init__(n_dims, template_dir=template_dir)
        self._cache = {lm_id: FileTemplateAdapter.load_template(self, lm_id)
                       for lm_id in FileTemplateAdapter.template_ids(self)}
        print('cached {} templates ({})'.format(
            len(self._cache), ', '.join(self._cache.keys())))

    def load_template(self, lm_id):
            return self._cache[lm_id]



import json
import os
import os.path as p
from os.path import expanduser, abspath
import shutil
import gzip
import struct
from functools import partial
from pathlib import Path
import time

import menpo
import menpo3d
from menpo.shape.mesh import TexturedTriMesh
import numpy as np

from landmarkerio import CacheFile


def filename_as_asset_id(fp):
    # simply return the filename as the asset_id
    return Path(fp).stem


def filepath_as_asset_id_under_dir(asset_dir):
    # find the filepath under asset_dir and return the full path as an asset id
    asset_dir = Path(abspath(expanduser(asset_dir)))

    def path_as_asset_id(fp):
        return '__'.join(Path(fp).relative_to(asset_dir).parts)

    return path_as_asset_id


def build_asset_mapping(identifier_f, asset_paths_iter):
    asset_mapping = {}
    for path in asset_paths_iter:
        asset_id = identifier_f(path)
        if asset_id in asset_mapping:
            raise RuntimeError(
                "asset_id {} is not unique - links to {} and "
                "{}".format(asset_id, asset_mapping[asset_id], path))
        asset_mapping[asset_id] = path
    return asset_mapping


# ASSET PATH RESOLUTION

def asset_paths(path_f, asset_dir, glob_pattern):
    return path_f(p.join(asset_dir, glob_pattern))

mesh_paths = partial(asset_paths, menpo3d.io.mesh_paths)
image_paths = partial(asset_paths, menpo.io.image_paths)


def glob_pattern(ext_str, recursive):
    file_glob = '*' + ext_str
    if recursive:
        return os.path.join('**', file_glob)
    else:
        return file_glob


def ensure_asset_dir(asset_dir):
    asset_dir = p.abspath(p.expanduser(asset_dir))
    if not p.isdir(asset_dir):
        raise ValueError('{} is not a directory.'.format(asset_dir))
    print ('assets:    {}'.format(asset_dir))
    return asset_dir


# CACHING

def cache_asset(cache_dir, cache_f, path, asset_id):
    r"""
    Caches the info for a given asset id so it can be efficiently
    served in the future.

    Parameters
    ----------
    asset_id : `str`
    The id of the asset that needs to be cached
    """
    asset_cache_dir = p.join(cache_dir, asset_id)
    if not p.isdir(asset_cache_dir):
        os.mkdir(asset_cache_dir)
    cache_f(cache_dir, path, asset_id)


# IMAGE CACHING

def cache_image(cache_dir, path, asset_id):
    r"""Actually cache this asset_id.
    """
    img = menpo.io.import_image(path)
    _cache_image_for_id(cache_dir, asset_id, img)


def _cache_image_for_id(cache_dir, asset_id, img):
    asset_cache_dir = p.join(cache_dir, asset_id)
    image_info_path = p.join(asset_cache_dir, CacheFile.image)
    texture_path = p.join(asset_cache_dir, CacheFile.texture)
    thumbnail_path = p.join(asset_cache_dir, CacheFile.thumbnail)
    img_path = img.path

    # WebGL only allows textures of maximum dimension 4096
    ratio = 4096.0 / np.array(img.shape)
    if np.any(ratio < 1):
        # the largest axis of img will now be 4096
        img = img.rescale(ratio.min(), round='round')
        had_to_shrink = True
    else:
        had_to_shrink = False
    # 1. Save out the image metadata json
    image_info = {'width': img.width,
                  'height': img.height}
    with open(image_info_path, 'wb') as f:
        json.dump(image_info, f)

    # 2. Save out the image
    if img_path.suffix == '.jpg' and not had_to_shrink:
        # Original was a jpg that was suitable, save it
        shutil.copyfile(str(img_path), texture_path)
    else:
        # Original wasn't a jpg or was too big - make it so
        img.as_PILImage().save(texture_path, format='jpeg')
    # 3. Save out the thumbnail
    save_jpg_thumbnail_file(img, thumbnail_path)


def save_jpg_thumbnail_file(img, path, width=640):
    ip = img.as_PILImage()
    w, h = ip.size
    h2w = h * 1. / w
    ips = ip.resize((width, int(h2w * width)))
    ips.save(path, quality=20, format='jpeg')


# MESH CACHING

def cache_mesh(cache_dir, path, asset_id):
    mesh = menpo3d.io.import_mesh(path)
    if isinstance(mesh, TexturedTriMesh):
        _cache_image_for_id(cache_dir, asset_id, mesh.texture)
    _cache_mesh_for_id(cache_dir, asset_id, mesh)


def _cache_mesh_for_id(cache_dir, asset_id, mesh):
    asset_cache_dir = p.join(cache_dir, asset_id)
    mesh_tmp_path = p.join(asset_cache_dir, CacheFile.mesh_tmp)
    mesh_path = p.join(asset_cache_dir, CacheFile.mesh)
    # store out the raw file
    _export_raw_mesh(mesh_tmp_path, mesh)
    # compress the raw and remove the uncompressed
    f_in = open(mesh_tmp_path, 'rb')
    f_out = gzip.open(mesh_path, 'wb', compresslevel=1)
    f_out.writelines(f_in)
    f_out.close()
    f_in.close()
    os.unlink(mesh_tmp_path)


def _export_raw_mesh(path, m):
    normals = False  # for now we are just not exporting normals.
    is_textured = hasattr(m, 'tcoords')
    with open(str(path), 'wb') as f:
        f.write(struct.pack('IIII', m.n_tris, is_textured, normals, False))
        m.points[m.trilist].astype(np.float32).tofile(f)
        if normals:
            m.vertex_normals[m.trilist].astype(np.float32).tofile(f)
        if is_textured:
            m.tcoords.points[m.trilist].astype(np.float32).tofile(f)


def ensure_cache_dir(cache_dir):
    cache_dir = p.abspath(p.expanduser(cache_dir))
    if not p.isdir(cache_dir):
        print("Warning the cache dir does not exist - creating...")
        os.mkdir(cache_dir)
    print ('cache:     {}'.format(cache_dir))
    return cache_dir


def serial_cacher(cache, path_asset_id):
    for i, (path, asset_id) in enumerate(path_asset_id):
        print('Caching {}/{} - {}'.format(i + 1, len(path_asset_id), asset_id))
        cache(path, asset_id)


def parallel_cacher(cache, path_asset_id, n_jobs=-1):
    from joblib import Parallel, delayed
    Parallel(n_jobs=n_jobs, verbose=5)(delayed(cache)(path, asset_id)
                                       for path, asset_id in path_asset_id)


def build_cache(cacher_f, asset_path_f, cache_f, identifier_f, asset_dir,
                cache_dir, recursive=False, ext=None, glob=None):

    # 1. Ensure the asset_dir and cache_dir are present.
    asset_dir = ensure_asset_dir(asset_dir)
    cache_dir = ensure_cache_dir(cache_dir)

    if recursive:
        print('assets dir will be searched recursively.')

    if ext is not None:
        ext_str = '.' + ext
        print('only assets of type {} will be '
              'loaded.'.format(ext_str))
    else:
        ext_str = ''
    if glob is None:
        # Figure out the glob pattern and save it
        glob_ptn = glob_pattern(ext_str, recursive)
    else:
        glob_ptn = glob

    print('Using glob: "{}"'.format(glob_ptn))

    # Construct a mapping from id's to file paths
    asset_id_to_paths = build_asset_mapping(identifier_f,
                                            asset_path_f(asset_dir, glob_ptn))

    # Check cache for what needs to be updated
    asset_ids = set(asset_id_to_paths.iterkeys())
    cached = set(os.listdir(cache_dir))
    uncached = asset_ids - cached

    print('{} assets need to be added to '
          'the cache'.format(len(uncached)))
    cache = partial(cache_asset, cache_dir, cache_f)
    path_asset_id = [(asset_id_to_paths[a_id], a_id) for a_id in uncached]
    start = time.time()
    cacher_f(cache, path_asset_id)
    elapsed = time.time() - start
    if len(uncached) > 0:
        print('{} assets cached in {:.0f} seconds'.format(len(uncached),
                                                        elapsed))
    return cache_dir, asset_id_to_paths


build_mesh_serial_cache = partial(build_cache, serial_cacher, mesh_paths,
                                  cache_mesh)
build_image_serial_cache = partial(build_cache, serial_cacher, image_paths,
                                   cache_image)

build_mesh_parallel_cache = partial(build_cache, parallel_cacher, mesh_paths,
                                    cache_mesh)
build_image_parallel_cache = partial(build_cache, parallel_cacher, image_paths,
                                     cache_image)


def cache_assets(mode, identifier_f, asset_dir, cache_dir, recursive=False,
                 ext=None, glob=None):
    r"""

    """
    if mode == 'image':
        cache_builder = build_image_parallel_cache
    elif mode == 'mesh':
        cache_builder = build_mesh_parallel_cache
    else:
        raise ValueError("mode must be 'image' or 'mesh'")
    return cache_builder(identifier_f, asset_dir, cache_dir,
                         recursive=recursive, ext=ext, glob=glob)



import abc
import os
from pathlib import Path

from flask import safe_join
from landmarkerio import CacheFile


class ImageAdapter(object):
    r"""
    Abstract definition of an adapter that serves image assets
    """
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def image_info(self, asset_id):
        pass

    @abc.abstractmethod
    def texture_file(self, asset_id):
        pass

    @abc.abstractmethod
    def thumbnail_file(self, asset_id):
        pass

    @abc.abstractmethod
    def asset_ids(self):
        pass


class MeshAdapter(object):

    @abc.abstractmethod
    def asset_ids(self):
        pass

    @abc.abstractmethod
    def mesh(self, asset_id):
        pass


class CacheAdapter(object):

    def __init__(self, cache_dir):
        self.cache_dir = Path(os.path.abspath(os.path.expanduser(cache_dir)))


class ImageCacheAdapter(CacheAdapter, ImageAdapter):

    def __init__(self, cache_dir):
        CacheAdapter.__init__(self, cache_dir)
        self._image_asset_ids = [a.parent.name
                                 for a in self.cache_dir.glob(os.path.join('*',
                                                              CacheFile.image))
                                 if a.parent.parent == self.cache_dir]

    def image_info(self, asset_id):
        return reduce(safe_join,
                      (str(self.cache_dir), asset_id, CacheFile.image))

    def texture_file(self, asset_id):
        return reduce(safe_join, (str(self.cache_dir),
                                  asset_id, CacheFile.texture))

    def thumbnail_file(self, asset_id):
        return reduce(safe_join,
                      (str(self.cache_dir), asset_id, CacheFile.thumbnail))

    def asset_ids(self):
        return self._image_asset_ids


class MeshCacheAdapter(CacheAdapter, MeshAdapter):

    def __init__(self, cache_dir):
        CacheAdapter.__init__(self, cache_dir)
        self._mesh_asset_ids = [a.parent.name
                                for a in self.cache_dir.glob(os.path.join('*',
                                                             CacheFile.mesh))
                                if a.parent.parent == self.cache_dir]

    def mesh(self, asset_id):
        return reduce(safe_join, (str(self.cache_dir), asset_id,
                                  CacheFile.mesh))

    def asset_ids(self):
        return self._mesh_asset_ids



import abc
import json
import os.path as p
from pathlib import Path
import os
from collections import defaultdict
from flask import safe_join
import glob

from landmarkerio import LM_DIRNAME, FileExt


class LmAdapter(object):
    r"""
    Abstract definition of an adapter that can be passed to app_for_adapter in
    order to generate a legal Flask implementation of landmarker.io's REST API
    for Landmarks.
    """
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def asset_id_to_lm_id(self):
        pass

    @abc.abstractmethod
    def lm_ids(self, asset_id):
        r"""
        Return a list of lm_ids available for a given asset_id
        """
        pass

    @abc.abstractmethod
    def load_lm(self, asset_id, lm_id):
        pass

    @abc.abstractmethod
    def save_lm(self, asset_id, lm_id, lm_json):
        pass


class FileLmAdapter(LmAdapter):
    r"""
    Concrete implementation of LmAdapater that serves landmarks from the
    local filesystem.
    """

    def load_lm(self, asset_id, lm_id):
        fp = self.lm_fp(asset_id, lm_id)
        if not p.isfile(fp):
            raise IOError
        with open(fp, 'rb') as f:
            lm = json.load(f)
            return lm

    def save_lm(self, asset_id, lm_id, lm_json):
        r"""
        Persist a given landmark definition to disk.
        """
        fp = self.lm_fp(asset_id, lm_id)
        with open(fp, 'wb') as f:
            json.dump(lm_json, f, sort_keys=True, indent=4,
                      separators=(',', ': '))

    @abc.abstractmethod
    def lm_fp(self, asset_id, lm_id):
        # where a landmark should exist
        pass


class SeparateDirFileLmAdapter(FileLmAdapter):

    def __init__(self, lm_dir):
        if lm_dir is None:
            # By default place the landmarks in the cwd
            lm_dir = p.join(os.getcwd(), LM_DIRNAME)
        self.lm_dir = p.abspath(p.expanduser(lm_dir))
        if not p.isdir(self.lm_dir):
            print("Warning the landmark dir does not exist - creating...")
            os.mkdir(self.lm_dir)
        print('landmarks: {}'.format(self.lm_dir))

    def lm_fp(self, asset_id, lm_id):
        # where a landmark should exist
        return safe_join(safe_join(self.lm_dir, asset_id), lm_id + FileExt.lm)

    def lm_ids(self, asset_id):
        r"""
        Return
        """
        lm_files = self._lm_paths(asset_id=asset_id)
        return [p.splitext(p.split(f)[-1])[0] for f in lm_files]

    def asset_id_to_lm_id(self):
        r"""
        Return a dict mapping asset ID's to landmark IDs that are
        present on this server for that asset.
        """
        lm_files = self._lm_paths()
        mapping = defaultdict(list)
        for lm_path in lm_files:
            dir_path, filename = p.split(lm_path)
            lm_id = p.splitext(filename)[0]
            asset_id = p.split(dir_path)[1]
            mapping[asset_id].append(lm_id)
        return mapping

    def _lm_paths(self, asset_id=None):
        # what landmarks do exist and where
        if asset_id is None:
            asset_id = '*'
        g = glob.glob(p.join(safe_join(self.lm_dir, asset_id), '*'))
        return filter(lambda f: p.isfile(f) and
                                p.splitext(f)[-1] == FileExt.lm, g)

    def save_lm(self, asset_id, lm_id, lm_json):
        r"""
        Persist a given landmark definition to disk.
        """
        subject_dir = safe_join(self.lm_dir, asset_id)
        if not p.isdir(subject_dir):
            os.mkdir(subject_dir)
        super(SeparateDirFileLmAdapter, self).save_lm(asset_id, lm_id, lm_json)


class InplaceFileLmAdapter(FileLmAdapter):

    def __init__(self, asset_ids_to_paths):
        self.ids_to_paths = asset_ids_to_paths
        print('landmarks served inplace - found {} asset with '
              'landmarks'.format(len(self.asset_id_to_lm_id())))

    def lm_ids(self, asset_id):
        # always the same!
        if asset_id in self.ids_to_paths:
            return ['inplace']
        else:
            raise ValueError

    def asset_id_to_lm_id(self):
        r"""
        Return a dict mapping asset ID's to landmark IDs that are
        present on this server for that asset.
        """
        return {aid: ['inplace']
                for aid in self.ids_to_paths
                if self._lm_path_for_asset_id(aid).is_file()}

    def lm_fp(self, asset_id, lm_id):
        # note the lm_id is ignored. We just always return the .ljson file.
        return str(self._lm_path_for_asset_id(asset_id))

    def _lm_path_for_asset_id(self, asset_id):
        asset_path = Path(self.ids_to_paths[asset_id])
        return asset_path.with_suffix(FileExt.lm)



from functools import partial
from flask import Flask, request, send_file, make_response
from flask.ext.restful import abort, Api, Resource
#from flask.ext.restful.utils import cors
import cors  # until twilio/flask-restful/pull/276 is merged, see the package
from landmarkerio import Server, Endpoints, Mimetype

url = lambda *x: '/' + '/'.join(x)
asset = lambda f: partial(f, '<string:asset_id>')


def safe_send(x, fail_message):
    try:
        return x
    except Exception as e:
        print(e)
        return abort(404, message=fail_message)


def safe_send_file(mimetype, path, fail_message, gzip=False):
    try:
        r = make_response(send_file(path, mimetype=mimetype))
        if gzip:
            r.headers['Content-Encoding'] = 'gzip'
        return r
    except Exception as e:
        print(e)
        return abort(404, message=fail_message)

image_file = partial(safe_send_file, Mimetype.jpeg)
json_file = partial(safe_send_file, Mimetype.json)
gzip_json_file = partial(safe_send_file, Mimetype.json, gzip=True)
binary_file = partial(safe_send_file, Mimetype.binary)
gzip_binary_file = partial(binary_file, gzip=True)


def lmio_api(dev=False):
    r"""
    Generate a Flask App that will serve meshes landmarks and templates to
    landmarker.io

    Parameters
    ----------
    adapter: :class:`LandmarkerIOAdapter`
        Concrete implementation of the LandmarkerIOAdapter. Will be queried for
        all data to pass to landmarker.io.
    dev: `bool`, optional
        If True, listen to anyone for CORS.

    Returns
    -------
    api, app, api_endpoint
    """
    app = Flask(__name__)
    api = Api(app)

    cors_dict = {
        'origin': Server.origin,
        'headers': ['Origin', 'X-Requested-With', 'Content-Type', 'Accept'],
        'methods': ['HEAD', 'GET', 'POST', 'PATCH', 'PUT', 'OPTIONS', 'DELETE'],
        'credentials': True
    }

    if dev:
        # in development mode, accept CORS from anyone (but we can't use HTTPS)
        cors_dict['origin'] = '*'
        cors_dict['credentials'] = False
        app.debug = True

    api.decorators = [cors.crossdomain(**cors_dict)]
    return api, app


def add_mode_endpoint(api, mode):

    if mode not in ['image', 'mesh']:
        raise ValueError("Mode can only be 'image' or 'mesh', "
                         "not {}".format(mode))

    class Mode(Resource):

        def get(self):
            return mode

    api.add_resource(Mode, url(Endpoints.mode))


def add_lm_endpoints(api, lm_adapter, template_adapter):
    r"""
    Generate a Flask App that will serve meshes landmarks and templates to
    landmarker.io

    Parameters
    ----------
    adapter: :class:`LandmarkerIOAdapter`
        Concrete implementation of the LandmarkerIOAdapter. Will be queried for
        all data to pass to landmarker.io.
    dev: `bool`, optional
        If True, listen to anyone for CORS.

    Returns
    -------
    api, app, api_endpoint
    """

    class Landmark(Resource):

        def get(self, asset_id, lm_id):
            err = "{} does not have {} landmarks".format(asset_id, lm_id)
            try:
                return lm_adapter.load_lm(asset_id, lm_id)
            except Exception as e:
                try:
                    return template_adapter.load_template(lm_id)
                except Exception as e:
                    return abort(404, message=err)

        def put(self, asset_id, lm_id):
            try:
                return lm_adapter.save_lm(asset_id, lm_id, request.json)
            except Exception as e:
                print(e)
                return abort(409, message="{}:{} unable to "
                                          "save".format(asset_id, lm_id))

        # Need this here to enable CORS put see http://mzl.la/1rCDkWX
        def options(self, asset_id, lm_id):
            pass

    class LandmarkList(Resource):

        def get(self):
            return lm_adapter.asset_id_to_lm_id()

    class LandmarkListForId(Resource):

        def get(self, asset_id):
            return lm_adapter.lm_ids(asset_id)

    lm_url = partial(url, Endpoints.landmarks)
    api.add_resource(LandmarkList, lm_url())
    api.add_resource(LandmarkListForId, asset(lm_url)())
    api.add_resource(Landmark, asset(lm_url)('<string:lm_id>'))


def add_template_endpoints(api, adapter):

    class Template(Resource):

        def get(self, lm_id):
            err = "{} template does not exist".format(lm_id)
            return safe_send(adapter.load_template(lm_id), err)

    class TemplateList(Resource):

        def get(self):
            return adapter.template_ids()

    templates_url = partial(url, Endpoints.templates)
    api.add_resource(TemplateList, templates_url())
    api.add_resource(Template, templates_url('<string:lm_id>'))


def add_collection_endpoints(api, adapter):

    class Collection(Resource):

        def get(self, collection_id):
            err = "{} collection not exist".format(collection_id)
            return safe_send(adapter.collection(collection_id), err)

    class CollectionList(Resource):

        def get(self):
            return adapter.collection_ids()

    collections_url = partial(url, Endpoints.collections)
    api.add_resource(CollectionList, collections_url())
    api.add_resource(Collection, collections_url('<string:collection_id>'))


def add_image_endpoints(api, adapter):
    r"""
    Generate a Flask App that will serve images, landmarks and templates to
    landmarker.io

    Parameters
    ----------

    adapter: :class:`ImageLandmarkerIOAdapter`
        Concrete implementation of the Image adapter. Will be queried for
        all data to pass to landmarker.io.

    """
    class Image(Resource):

        def get(self, asset_id):
            err = "{} does not have an image".format(asset_id)
            return json_file(adapter.image_info(asset_id), err)

    class ImageList(Resource):

        def get(self):
            return adapter.asset_ids()

    class Texture(Resource):

        def get(self, asset_id):
            err = "{} does not have a texture".format(asset_id)
            return image_file(adapter.texture_file(asset_id), err)

    class Thumbnail(Resource):

        def get(self, asset_id):
            err = "{} does not have a thumbnail".format(asset_id)
            return image_file(adapter.thumbnail_file(asset_id), err)

    image_url = partial(url, Endpoints.images)
    texture_url = partial(url, Endpoints.textures)
    thumbnail_url = partial(url, Endpoints.thumbnail)

    api.add_resource(ImageList, image_url())
    api.add_resource(Image, asset(image_url)())

    api.add_resource(Texture, asset(texture_url)())
    api.add_resource(Thumbnail, asset(thumbnail_url)())


def add_mesh_endpoints(api, adapter):

    class Mesh(Resource):

        def get(self, asset_id):
            err = "{} is not an available mesh".format(asset_id)
            return gzip_binary_file(adapter.mesh(asset_id), err)

    class MeshList(Resource):

        def get(self):
            return adapter.asset_ids()

    mesh_url = partial(url, Endpoints.meshes)
    mesh_asset_url = asset(mesh_url)

    api.add_resource(MeshList, mesh_url())
    api.add_resource(Mesh, mesh_asset_url())



from landmarkerio.server import (lmio_api, add_mode_endpoint, add_lm_endpoints,
                                 add_image_endpoints, add_mesh_endpoints,
                                 add_template_endpoints,
                                 add_collection_endpoints)
from landmarkerio.template import CachedFileTemplateAdapter
from landmarkerio.collection import (AllCacheCollectionAdapter,
                                     FileCollectionAdapter)
from landmarkerio.asset import ImageCacheAdapter, MeshCacheAdapter

from landmarkerio import Server


def serve_with_cherrypy(app, port=5000, public=False):
    import cherrypy
    # Mount the WSGI callable object (app) on the desired endpoint (e.g.
    # /api/v1)
    cherrypy.tree.graft(app, Server.endpoint)
    update_dict = {
        'server.socket_port': port,
    }
    if public:
        update_dict['server.socket_host'] = '0.0.0.0'

    cherrypy.config.update(update_dict)
    # Start the CherryPy WSGI web server
    cherrypy.engine.start()
    cherrypy.engine.block()


def serve_from_cache(mode, cache_dir, lm_adapter, template_dir=None,
                     collection_dir=None, dev=False):
    r"""

    """
    api, app = lmio_api(dev=dev)
    if dev:
        app.debug = True
    # always serve at least images
    add_image_endpoints(api, ImageCacheAdapter(cache_dir))
    if mode == 'image':
        n_dims = 2
    elif mode == 'mesh':
        n_dims = 3
        add_mesh_endpoints(api, MeshCacheAdapter(cache_dir))
    else:
        raise ValueError("mode must be 'image' or 'mesh'")
    add_mode_endpoint(api, mode)
    template_adapter = CachedFileTemplateAdapter(n_dims,
                                                 template_dir=template_dir)
    add_template_endpoints(api, template_adapter)
    add_lm_endpoints(api, lm_adapter, template_adapter)
    if collection_dir is not None:
        collection_adapter = FileCollectionAdapter(collection_dir)
    else:
        collection_adapter = AllCacheCollectionAdapter(cache_dir)
    add_collection_endpoints(api, collection_adapter)
    print(collection_adapter)
    return app
